if(massiveextender)
{
	let $ns = 		massiveextender.REQUIRE(massiveextender.Helpers.NS);
	let $dom =		massiveextender.REQUIRE(massiveextender.Helpers.DOM);
	let $xre = 		massiveextender.REQUIRE(massiveextender.Helpers.XULRT);
	let $xpc =		massiveextender.REQUIRE(massiveextender.Helpers.XPC);
	let $u =		massiveextender.REQUIRE(massiveextender.Helpers.Utilities);

	if(!massiveextender.App) massiveextender.App = {};
	
	// Persistent storage service
	// TODO (future): ideally reform this into a "framework" namespace which provides
	// the creation of "service instances" with semi-automatic interface for DB creation/management
	// and data retrieval ("microstorage"? - similarly to the "microdom" concept)
	// TODO (medium): transactions
	let $db =
	massiveextender.App.Persist =
	{
		InitDB: function()
		{
			var connection = $db.connection;
			if(!connection.tableExists("collections"))
				connection.createTable("collections",
						'"ID" INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT, '+
						'"bwurl" TEXT NOT NULL UNIQUE');
			if(!connection.tableExists("addons"))
				connection.createTable("addons",
						'"ID" INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT, '+
						'"guid" TEXT NOT NULL UNIQUE');
			if(!connection.tableExists("exceptions"))
				connection.createTable("exceptions",
						'"ID" INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT, '+
						'"collectionID" INTEGER NOT NULL, "addonID" INTEGER NOT NULL');
			var sqlstr, statement;
			try
			{
				sqlstr = 'CREATE INDEX IF NOT EXISTS "collections_bwurl_index" ON "collections" ("bwurl")';
				statement = connection.createStatement(sqlstr);
				statement.execute();
				sqlstr = 'CREATE INDEX IF NOT EXISTS "addons_guid_index" ON "addons" ("guid")';
				statement = connection.createStatement(sqlstr);
				statement.execute();
			}
			catch(ex) { } // not critical (TODO (low): keep an eye on this)
			finally
			{
				statement.reset();
			}
		}, // InitDB
		
		CloseDB: function()
		{
			// TODO (medium): cleanup (to be performed at ext.manager shutdown)
			$db.connection.close();
		}, // CloseDB
		
		
		// -- utilities --

		_connection: null,
		get connection()
		{
			if(!$db._connection)
				$db._connection =
						$xpc.service.byid("@mozilla.org/storage/service;1","mozIStorageService").
								openDatabase($db.file);
			return $db._connection;
		}, // get connection
		
		_file: null,
		get file()
		{
			if(!$db._file)
			{
				$db._file = $xre.ProfileDir;
				$db._file.append('total-collector.sqlite');
			}
			return $db._file;
		}, // get file
		
		/**
		* Binds parameter to statement
		* @param	statement
		* 		mozIStorageStatement
		* @param	paramnumber
		* 		Either a number representing 0-based API index, or string representing 1-based SQL index
		* @param	value
		* 		Value to be bound. Can be omitted to bind null. Otherwise numbers bound to numbers,
		* 		booleans to number 1/0, anything other bound to string
		**/
		Bind: function(statement, paramnumber, value)
		{
			if(typeof paramnumber == 'string')
			{
				paramnumber = Number(paramnumber);
				paramnumber--;
			}
			if (!value)
				statement.bindNullParameter(paramnumber);
			else
			{
				switch(typeof value)
				{
					case 'number':
						statement.bindInt32Parameter(paramnumber, value);
						break;
					case 'boolean':
						statement.bindInt32Parameter(paramnumber, value ? 1 : 0);
						break;
					default:
						statement.bindStringParameter(paramnumber, value);
						break;
				}
			}
		}, // Bind
		
		Extract: function(sqlstatement, columnname)
		{
			var index = sqlstatement.getColumnIndex(columnname);
			try
			{
				var sqltype = sqlstatement.getTypeOfIndex(index);
				switch (sqltype)
				{
					case sqlstatement.VALUE_TYPE_INTEGER:
						return sqlstatement.getInt32(index);
					case sqlstatement.VALUE_TYPE_FLOAT:
						return sqlstatement.getDouble(index);
					case sqlstatement.VALUE_TYPE_TEXT:
						return sqlstatement.getString(index);
					case sqlstatement.VALUE_TYPE_BLOB:
						return sqlstatement.getBlob(index);
					case sqlstatement.VALUE_TYPE_NULL:
					default:
						return null;
				}
			}
			catch(ex)
			{
				return null;
			}
		}, // Extract

		// -- service interface --
		
		FindCollectionID: function(bwurl)
		{
			var sqlresult = null;
			var sqlstr =
					'SELECT "ID" FROM "collections" '+
					'WHERE ("bwurl" = ?1) LIMIT 1';
			var statement;
			try
			{
				statement = $db.connection.createStatement(sqlstr);
				$db.Bind(statement, '1', bwurl);
				statement.executeStep()
				sqlresult = $db.Extract(statement, 'ID');
			}
			catch(ex)
			{
				throw(ex);
			}
			finally
			{
				statement.reset();
			}
			return (sqlresult == null ? -1 : sqlresult);
		}, // FindCollectionID
		FindAddonID: function(guid)
		{
			var sqlresult = null;
			var sqlstr =
					'SELECT "ID" FROM "addons" '+
					'WHERE ("guid" = ?1) LIMIT 1';
			var statement;
			try
			{
				statement = $db.connection.createStatement(sqlstr);
				$db.Bind(statement, '1', guid);
				statement.executeStep()
				sqlresult = $db.Extract(statement, 'ID');
			}
			catch(ex)
			{
				throw(ex);
			}
			finally
			{
				statement.reset();
			}
			return (sqlresult == null ? -1 : sqlresult);
		}, // FindAddonID
		StoreNewCollection: function(bwurl)
		{
			var success;
			var sqlstr =
					'INSERT INTO "collections" '+
					'("bwurl") VALUES (?1)';
			var statement;
			try
			{
				statement = $db.connection.createStatement(sqlstr);
				$db.Bind(statement, '1', bwurl);
				statement.execute();
				success = true;
			}
			catch(ex)
			{
				success = false;
				//throw(ex);
			}
			finally
			{
				statement.reset();
			}
			return success;
		}, // StoreNewCollection
		StoreNewAddon: function(guid)
		{
			var success;
			var sqlstr =
					'INSERT INTO "addons" '+
					'("guid") VALUES (?1)';
			var statement;
			try
			{
				statement = $db.connection.createStatement(sqlstr);
				$db.Bind(statement, '1', guid);
				statement.execute();
				success = true;
			}
			catch(ex)
			{
				success = false;
				//throw(ex);
			}
			finally
			{
				statement.reset();
			}
			return success;
		}, // StoreNewAddon
		StoreExceptions: function(collectionID, addonIDs)
		{
			var success;
			var sqlstr =
					'INSERT OR REPLACE INTO "exceptions" '+
					'("collectionID", "addonID") VALUES (?1, ?2)';
			var statement;
			try
			{
				$db.connection.beginTransaction();
				for(let addonIDnr = 0, addoncount = addonIDs.length; addonIDnr < addoncount; addonIDnr++)
				{
					statement = $db.connection.createStatement(sqlstr);
					$db.Bind(statement, '1', collectionID);
					$db.Bind(statement, '2', addonIDs[addonIDnr]);
					statement.execute();
				}
				$db.connection.commitTransaction();
				success = true;
			}
			catch(ex)
			{
				$db.connection.rollbackTransaction();
				success = false;
				//throw(ex);
			}
			finally
			{
				statement.reset();
			}
			return success;
		}, // StoreException
		RemoveExceptions: function(collectionID, addonIDs)
		{
			var success;
			var sqlstr =
					'DELETE FROM "exceptions" '+
					'WHERE "collectionID" = ?1 AND "addonID" = ?2';
			var statement;
			try
			{
				$db.connection.beginTransaction();
				for(let addonIDnr = 0, addoncount = addonIDs.length; addonIDnr < addoncount; addonIDnr++)
				{
					statement = $db.connection.createStatement(sqlstr);
					$db.Bind(statement, '1', collectionID);
					$db.Bind(statement, '2', addonIDs[addonIDnr]);
					statement.executeStep();
				}
				$db.connection.commitTransaction();
				success = true;
			}
			catch(ex)
			{
				$db.connection.rollbackTransaction();
				success = false;
				//throw(ex);
			}
			finally
			{
				statement.reset();
			}
			return success;
		}, // RemoveException
		GetException: function(collectionID, addonID)
		{
			var sqlresult = null;
			var sqlstr =
					'SELECT "ID" FROM "exceptions" '+
					'WHERE "collectionID" = ?1 AND "addonID" = ?2 LIMIT 1';
			var statement;
			try
			{
				statement = $db.connection.createStatement(sqlstr);
				$db.Bind(statement, '1', collectionID);
				$db.Bind(statement, '2', addonID);
				statement.executeStep();
				sqlresult = $db.Extract(statement, 'ID');
			}
			catch(ex)
			{
				throw(ex);
			}
			finally
			{
				statement.reset();
			}
			return (sqlresult != null);
		} // GetException
	} // massiveextender.App["Persist"]

	let $eng =
	massiveextender.App.Engine =
	{
		// Add-on Collector glue methods
		// TODO (low): the right thing to do would probably be to instantiate a collection object
		// when collection is changed or updated in A.C.
		get selectedcollection()
		{
			return Bandwagon.Controller.CollectionsPane.
					elemBandwagonCollections.selectedItem;
		},
		get selectedaddon()
		{
			return Bandwagon.Controller.ExtensionsOverlay._getSelectedExtension();
		},
		
		collection:
		{
			get name()
			{
				return $eng.selectedcollection.collection.name;
			},
			get bwurl()
			{
				return $eng.selectedcollection.collection.resourceURL;
			},
			get addons()
			{
				if($eng.selectedcollection.collection.hasAddon())
					return $eng.selectedcollection.collection.addons;
				return null;
			},
			// TODO (med): optimize this getter with a cached value; will need to track
			// collection selection change
			get size()
			{
				let size = 0;
				let addons = $eng.collection.addons;
				if(addons)
					for(let addon in addons) // this is not an array
						size++;
				return size;
			},
			// TODO (med): optimize this getter with a cached value; will need to track
			// collection selection change
			get selectionsize()
			{
				let selectionsize = 0;
				let addons = $eng.collection.addons;
				if(addons)
					for each(let addon in addons) // this is not an array
						if(!$eng.IsExcluded($eng.collection.bwurl, addon.guid))
							selectionsize++;
				return selectionsize;
			}
		},

		// Storage interface glue methods
		// TODO (low): the right thing to do would probably be to instantiate an addon object
		IsExcluded: function(bwurl, guid)
		{
			let addonID = $db.FindAddonID(guid);
			if(addonID < 0)
				return false;
			let collectionID = $db.FindCollectionID(bwurl);
			if(collectionID < 0)
				return false;
			return $db.GetException(collectionID, addonID);
		},
		StoreExceptions: function(bwurl, guids)
		{
			if(guids.length == 0)
				return;
			let collectionID = $db.FindCollectionID(bwurl);
			if(collectionID < 0)
			{
				$db.StoreNewCollection(bwurl);
				collectionID = $db.FindCollectionID(bwurl);
			}
			let addonIDs = [];
			for(guidnr = 0, count = guids.length; guidnr < count; guidnr++)
			{
				if(!guids[guidnr])
					continue;
				let addonID = $db.FindAddonID(guids[guidnr]);
				if(addonID < 0)
				{
					$db.StoreNewAddon(guids[guidnr]);
					addonID = $db.FindAddonID(guids[guidnr]);
				}
				addonIDs.push(addonID);
			}
			if(addonIDs.length == 0)
				return;
			$db.StoreExceptions(collectionID, addonIDs);
		},
		RemoveExceptions: function(bwurl, guids)
		{
			if(guids.length == 0)
				return;
			let collectionID = $db.FindCollectionID(bwurl);
			// if collection is not in database, we simply return
			if(collectionID < 0)
				return;
			let addonIDs = [];
			for(guidnr = 0, count = guids.length; guidnr < count; guidnr++)
			{
				if(!guids[guidnr])
					continue;
				let addonID = $db.FindAddonID(guids[guidnr]);
				if(addonID >= 0)
					addonIDs.push(addonID);
			}
			if(addonIDs.length == 0)
				return;
			$db.RemoveExceptions(collectionID, addonIDs);
		},

		// Extension Manager glue methods
		extmanager:
		{
			_extmansvc : null,
			get extmansvc()
			{
				if(!$eng.extmanager._extmansvc)
					$eng.extmanager._extmansvc =
							$xpc.service.byid("@mozilla.org/extensions/manager;1", "nsIExtensionManager");
				return $eng.extmanager._extmansvc;
			},
			IsInstalled: function(guid)
			{
				return (this.extmansvc.getItemForID(guid)) ? true : false;
			},
			IsEnabled: function(guid)
			{
				if(!this.IsInstalled(guid))
					return false;
				let rdfsvc = $xpc.service.byid("@mozilla.org/rdf/rdf-service;1", "nsIRDFService");
				let src = rdfsvc.GetResource("urn:mozilla:item:" + guid);
				let link = rdfsvc.GetResource($ns.mozEM + 'Disabled');
				let ds = this.extmansvc.datasource;
				let target = ds.GetTarget(src, link, true);
				if( (target instanceof $xpc.i["nsIRDFLiteral"]) &&
						(target.Value == "true") )
					return false;
				link = rdfsvc.GetResource($ns.mozEM + 'opType');
				// TODO (high): check to see if the following wrongly reports pending updates and installs as disabled
				target = ds.GetTarget(src, link, false)
				return target ? false : true;
			},
			Enable: function(guid)
			{
				if(this.extmansvc.getItemForID(guid))
					if(this.extmansvc.getInstallLocation(guid))
						this.extmansvc.enableItem(guid);
			},
			Disable: function(guid)
			{
				if(this.extmansvc.getItemForID(guid))
					if(this.extmansvc.getInstallLocation(guid))
						this.extmansvc.disableItem(guid);
			},
			Uninstall: function(guid)
			{
				if(this.extmansvc.getItemForID(guid))
				{
					let location = this.extmansvc.getInstallLocation(guid);
					if(!location)
						return;
					if(location.canAccess)
						this.extmansvc.uninstallItem(guid);
					else
						this.extmansvc.disableItem(guid);
				}
			},
			ExtensionDir: function(guid)
			{
				// the path may use forward slash ("/") as the delimiter  
				// returns nsIFile for the extension's install.rdf  
				return this.extmansvc.getInstallLocation(guid).
						getItemFile(guid, "install.rdf");  
			},

			IsOpPending: function(guid)
			{
				if(!this.IsInstalled(guid))
					return null;
				let rdfsvc = $xpc.service.byid("@mozilla.org/rdf/rdf-service;1", "nsIRDFService");
				let src = rdfsvc.GetResource("urn:mozilla:item:" + guid);
				let link = rdfsvc.GetResource($ns.mozEM + 'Disabled');
				let ds = this.extmansvc.datasource;
				link = rdfsvc.GetResource($ns.mozEM + 'opType');
				let target = ds.GetTarget(src, link, false)
				return target ? false : true;
			},
			/**
			* Installs a batch using InstallTrigger, prompting EULA acceptance for all items
			* that require this
			* @param		istruct
			* 			batch installation structure returned by ObtainInstallStruct()
			* 			(only .eulas and .xpis used, no option processing!)
			* @param		callback
			* 			InstallTrigger callback
			*/
			InstallBatch: function(istruct, callback)
			{
				if(istruct.eulas)
				{
					for(let name in istruct.eulas)
					{
						let eula = istruct.eulas[name];
						if(!eula)
							continue;
						with($ui.windows["eula"])				
							window.openDialog(xuluri,name,flags,eula);
						if (!eula.accepted)
						{
							delete istruct.xpis[name];
							if(istruct.addons)
								delete istruct.addons[name]
						}
					}
				}
				
				let xpisempty = true;
				for(let name in istruct.xpis)
				{
					xpisempty = false;
					break;
				}	
				if(xpisempty)
					return false;
				try
				{
					InstallTrigger.install(istruct.xpis, callback)
				}
				catch(ex)
				{
					// TODO (low): log this to ensure we didn't omit some faulty scenarios
					return false; 
				} 
				return true;
			}
		},
		
		/**
		* Prepares install structure for the requested batch
		* that require this
		* @param		howmany
		* 			either "selected" or "all"
		* @returns	istruct, including counters, addon data, eulas and xpis suitable for
		* 			InstallTrigger
		*/
		// TODO (low): if the whole extension management chunk is going to be spun off into a
		// generic service, this one can go as well provided we feed it with an iterator
		// instead of directly accessing A.C. related stuff
		// TODO (high): need to generalize this with high priority to manage bookmarks stuff too
		ObtainInstallStruct: function(howmany)
		{
			if(!$eng.selectedcollection)
				return null;
			let addons = $eng.collection.addons;
			if(!addons)
				return null;
				
			var istruct =
			{
				collection: {},
				addons: {},
				xpis: {},
				eulas: {},
				
				options:
				{
					"force update": false,
					"enable disabled": false,
					"install experimental": false,
					"allow incompat": false
				},
				counters:
				{
					total: 0,
					actionable: 0,
					"installed":
					{
						"enabled": 0,
						"disabled": 0
					},
					"missing":
					{
						"installable": 0,
						"experimental": 0,
						"incompatible": 0,
						"refuse": 0
					}
				},
				// we need to drag this enum with the whole struct in order to pass it to
				// batch options/confirmation dialog
				itype:
				{
					"installable": 			Bandwagon.Model.Addon.INSTALL_YES,
					"needs downgrade": 		Bandwagon.Model.Addon.INSTALL_NO_ADDON_IS_FOR_OLDER_VERSION,
					"needs upgrade": 		Bandwagon.Model.Addon.INSTALL_NO_UPGRADE_TO_USE_THIS_VERSION,
					"needs beta": 			Bandwagon.Model.Addon.INSTALL_NO_MUST_DOWNLOAD_BETA,
					"wrong os": 			Bandwagon.Model.Addon.INSTALL_NO_NOT_COMPATIBLE_OS,
					"experimental": 		Bandwagon.Model.Addon.INSTALL_YES_IS_EXPERIMENTAL,
					"already installed": 	Bandwagon.Model.Addon.INSTALL_NO_ALREADY_INSTALLED
				}
			};

			istruct.collection.name = $eng.collection.name;

			for each(let addon in addons)
			{
				if($eng.IsExcluded($eng.collection.bwurl, addon.guid) && (howmany == "selected"))
					continue;
				let name = addon.name;
				istruct.addons[name] = {};
				istruct.addons[name].installer = 
						addon.getInstaller(
								Bandwagon.Util.getHostEnvironmentInfo().os);
				istruct.addons[name].guid = addon.guid;
				istruct.addons[name].itype =
						addon.canInstall(Bandwagon.Util.getHostEnvironmentInfo()).
						type;
			
				switch(istruct.addons[name].itype)
				{
				case istruct.itype["already installed"]:
					if($eng.extmanager.IsEnabled(istruct.addons[name].guid))
					{
						istruct.addons[name].isenabled = true;
						istruct.counters["installed"]["enabled"]++;
					}
					else
						istruct.counters["installed"]["disabled"]++;
					break;
				case istruct.itype["experimental"]:
					istruct.counters["missing"]["experimental"]++;
					break;
				case istruct.itype["installable"]:
					istruct.counters["missing"]["installable"]++;
					break;
				case istruct.itype["wrong os"]:
					istruct.counters["missing"]["refuse"]++;
					break;
				default:
					istruct.counters["missing"]["incompatible"]++;
				}
				istruct.counters.total++;
				if( (!$eng.extmanager.IsInstalled(istruct.addons[name].guid)) &&
						(istruct.addons[name].itype != istruct.itype["wrong os"]) )
				{
					if(addon.eula && addon.eula != "")
						istruct.eulas[name] =
						{
							name: addon.name,
							text: addon.eula,
							accepted: false
						}
					istruct.xpis[name] = istruct.addons[name].installer;
				}
			}
			return (istruct.counters.total > 0) ? istruct : null;
		}, // ObtainInstallStruct
		
		InitiateInstallation: function(istruct)
		{
			if(istruct.counters.actionable == 0) // TODO (low): redundant check, needs to go
				return false;
			// TODO (low): see comments on needrestart at end of procedure
			let needrestart = { enable: false, install: false };
			let itype = istruct.itype;
			let options = istruct.options;
			let addons = istruct.addons;
			for(let name in addons)
			{
				let addon = addons[name];
				switch(addon.itype)
				{
				case itype["already installed"]:
					if( options["enable disabled"] &&
							(!addon.isenabled) )
					{
						$eng.extmanager.Enable(addons[name].guid);
						needrestart.enable = true;
					}
					break;
				case itype["experimental"]:
					if(!options["install experimental"])
					{
						delete istruct.xpis[name];
						delete istruct.eulas[name];
					}
					break;
				case itype["needs downgrade"]:
				case itype["needs upgrade"]:
				case itype["needs beta"]:
					if(!options["allow incompat"])
					{
						delete istruct.xpis[name];
						delete istruct.eulas[name];
					}
					break;
				}
			}
			// TODO (low): this is a lame job - ideally we need to supply callback to InstallBatch
			// otherwise we'll be tricked into thinking we need restart even if xpi installation
			// completely fails (e.g. due to network out); however since we finally let E.M.
			// decide, this probably is of low concern (TODO: need to test; however if this is the
			// case then this whole needrestart thing is bogus and possibly needs to go)
			needrestart.install = $eng.extmanager.InstallBatch(istruct);
			return needrestart.enable || needrestart.install;
		} // InitiateInstallation

	} // massiveextender.App["Engine"]
	
	let $ui =
	massiveextender.App.UI = 
	{
		// TODO (high): display notification box after a batch job has been carried out (especially installation)
		commands:
		{
			IncludeInBatch: function()
			{
				if($ui._inclusion_checkbox.getAttribute("checked") == "true")
					with($eng) { RemoveExceptions(collection.bwurl, [selectedaddon.guid]); }
				else
					with($eng) { StoreExceptions(collection.bwurl, [selectedaddon.guid]); }
				$ui.SetCheckBoxState();
				$ui.StylizeBandwagonAddons();
				$ui.ReportSelection();
			},
			Select: function(howmany)
			{
				if(!$eng.selectedcollection)
					return;
				let addons = $eng.collection.addons;
				if(!addons)
					return;
				let bwurl = $eng.collection.bwurl;
				let guidsforremoval = [], guidsforstoring = [];
				let map = function(addons, lambda)
				{
					let result = [];
					for each(addon in addons)
						result.push(lambda(addon))
					return result;
				}
				switch(howmany)
				{
				case "all":
					guidsforremoval = map(addons, function(addon) (addon.guid));
					break;
				case "none":
					guidsforstoring = map(addons, function(addon) (addon.guid));
					break;
				case "flip":
					guidsforremoval = map(addons, function(addon) ($eng.IsExcluded(bwurl, addon.guid) ? addon.guid : null));
					guidsforstoring = map(addons, function(addon) ($eng.IsExcluded(bwurl, addon.guid) ? null : addon.guid));
					break;
				case "-installed":
/*					let filterinstalled = function(addon)
					{
						let bwinstallable = addon.canInstall(Bandwagon.Util.getHostEnvironmentInfo()).type;
						return (bwinstallable == Bandwagon.Model.Addon.INSTALL_NO_ALREADY_INSTALLED) ? addon.guid : null;
					}
					guidsforstoring = map(addons, filterinstalled);
*/					let filterinstalled = function(addon)
						let(bwinstallable = addon.canInstall(Bandwagon.Util.getHostEnvironmentInfo()).type)
							(bwinstallable == Bandwagon.Model.Addon.INSTALL_NO_ALREADY_INSTALLED) ? addon.guid : null;
					guidsforstoring = map(addons, filterinstalled);
					break;
				case "-incompatible":
					let filterincompatible = function(addon)
						let(bwinstallable = addon.canInstall(Bandwagon.Util.getHostEnvironmentInfo()).type)
							( (bwinstallable != Bandwagon.Model.Addon.INSTALL_YES) &&
							(bwinstallable != Bandwagon.Model.Addon.INSTALL_YES_IS_EXPERIMENTAL) &&
							(bwinstallable != Bandwagon.Model.Addon.INSTALL_NO_ALREADY_INSTALLED) ) ?
							addon.guid : null;
					guidsforstoring = map(addons, filterincompatible);
					break;
				}
				$eng.RemoveExceptions(bwurl, guidsforremoval);
				$eng.StoreExceptions(bwurl, guidsforstoring);

				$ui.SetCheckBoxState();
				$ui.StylizeBandwagonAddons();
				$ui.ReportSelection();

			},
			Install: function(howmany)
			{
				let istruct = $eng.ObtainInstallStruct(howmany);
				if(!istruct)
					return;

				/* --- CALLING OPTIONS AND CONFIRMATION DIALOG --- */
				with($ui.windows["confirminstall"])				
					window.openDialog(xuluri,name,flags,istruct);
				/* ----------------------------------------------- */

				if((typeof(istruct.go)=='undefined') || (!istruct.go))
					return;

				if($eng.InitiateInstallation(istruct))
					$ui.UpdateRestartStatus("noviewswitch");
			},
			Uninstall: function(howmany)
			{
				if(!$eng.selectedcollection)
					return;
				let addons = $eng.collection.addons;
				if(!addons)
					return;
				let needrestart = false;
				for each(let addon in addons)
					if($eng.IsExcluded($eng.collection.bwurl, addon.guid) && (howmany == "selected"))
						continue;
					else
					{
						$eng.extmanager.Uninstall(addon.guid);
						needrestart = true;
					}
				if(needrestart)
					$ui.UpdateRestartStatus();
			},
			Enable: function(howmany)
			{
				if(!$eng.selectedcollection)
					return;
				let addons = $eng.collection.addons;
				if(!addons)
					return;
				let needrestart = false;
				for each(let addon in addons)
					if($eng.IsExcluded($eng.collection.bwurl, addon.guid) && (howmany == "selected"))
						continue;
					else
					{
						$eng.extmanager.Enable(addon.guid);
						needrestart = true;
					}
				if(needrestart)
					$ui.UpdateRestartStatus();
			},
			Disable: function(howmany)
			{
				if(!$eng.selectedcollection)
					return;
				let addons = $eng.collection.addons;
				if(!addons)
					return;
				let needrestart = false;
				for each(let addon in addons)
					if($eng.IsExcluded($eng.collection.bwurl, addon.guid) && (howmany == "selected"))
						continue;
					else
					{
						$eng.extmanager.Disable(addon.guid);
						needrestart = true;
					}
				if(needrestart)
					$ui.UpdateRestartStatus();
			},
			SwitchUIMode: function(mode)
			{
				prefs = Application.extensions.get("{E6175356-15A3-11DF-AFEE-C56455D89593}").prefs;
				pref = "uimode";
				if(mode)
					prefs.setValue(pref, mode)
				else
				{
					if(prefs.has(pref))
						mode = prefs.get(pref).value
					else
					{
						mode = "simple";
						prefs.setValue(pref, mode);
					}
				}
			
				let advcmd = $dom.byid("massive-extender-switch-to-advanced-command");
				let smptbitem = $dom.byid("batch-actions-buttons-wrapper-toolbaritem");
				let	smpcmd = $dom.byid("massive-extender-switch-to-simple-command");
				let advtbitem = $dom.byid("batch-actions-menubar-wrapper-toolbaritem");
	
				advcmd.setAttribute("disabled", mode != "simple");
				smptbitem.setAttribute("hidden", mode != "simple");
				smpcmd.setAttribute("disabled", mode == "simple");
				advtbitem.setAttribute("hidden", mode == "simple")

			}
		},

		listeners:
		{
			BWAddons_select: function(ev)
			{
				$ui.InsertInclusionCheckBox();
			},

			ExtensionsView_select: function(ev)
			{
				$ui.InsertBookmarkButton();
			}
		},
		
		// TODO (medium): the idea here and in similar cases was to use "with" but it's better to switch
		// to self-contained OOP-style service instance
		windows:
		{
			confirminstall:
			{ 
				windowtype:		"massiveextender:ConfirmInstall", 
				xuluri:			"chrome://massive-extender/content/xul/dialog.install-options.xul",
				name:			"massiveextender-ConfirmInstall",
				flags:			"chrome,dialog=no,centerscreen,modal,resizable=no",
				openmultiple:	false
			},
			eula:
			{ 
				windowtype:		"", 
				xuluri:			"chrome://mozapps/content/extensions/eula.xul",
				name:			"_blank",
				flags:			"chrome,dialog,centerscreen,modal,resizable=no",
				openmultiple:	false
			}
		},

		// TODO (medium): ideally string bundles should be initialized at app load time by
		// the "framework" automatically, and possibly the best thing to do would be to join all
		// bundles required by a module into one - need to investigate if this is possible with XPCOM
		_stringbundle: null,
		get stringbundle()
		{
			if(!$ui._stringbundle)
				$ui._stringbundle = $dom.byid("massive-extender-stringbundle");
			return $ui._stringbundle;
		},
		
		/********************* ADD-ON COLLECTOR PLUGIN CODE *********************/ 
		_bwaddonmoreactions_box: null,
		_inclusion_checkbox: null,
		/**
		* Ensures that a checkbox-holding element is ready for insertion in
		* Add-on Collector add-on richlistbox
		* @returns 	Checkbox-holding <richlistboxitem>
		*/
		// TODO (medium): e4x templating looks better than manual dom manipulation but
		// still wants for more polish/automation
		PrepareBWAddonMoreActionsElement: function()
		{
			if (!$ui._bwaddonmoreactions_box)
			{
				let label = $ui.stringbundle.getString("inclusion.checkbox");
				
				let boxnodexml =
					<hbox xmlns={$ns["XUL"]} />;
				let spacerxml =
					<spacer xmlns={$ns["XUL"]} flex="1"/>;
				let checkboxxml =
					<checkbox xmlns={$ns["XUL"]} label={label} command="massive-extender-include-in-batch-command"/>
				
				$ui._bwaddonmoreactions_box = $dom.parse(boxnodexml);
				$ui._bwaddonmoreactions_box.appendChild($dom.parse(spacerxml));

				$ui._inclusion_checkbox = $dom.parse(checkboxxml);
				$ui._bwaddonmoreactions_box.appendChild($ui._inclusion_checkbox);
			}
			else
				if ($ui._bwaddonmoreactions_box.parentNode)
					$ui._bwaddonmoreactions_box.parentNode.
							removeChild($ui._bwaddonmoreactions_box);
			
			$ui.SetCheckBoxState();

			return $ui._bwaddonmoreactions_box;
		},
		
		/**
		* Inserts checkbox-holding element in A.C. add-on richlistbox
		*/
		InsertInclusionCheckBox: function()
		{
			let appendeditem = $ui.PrepareBWAddonMoreActionsElement();
			let addonsrichlist = $dom.byid("bandwagon-addons-list");
			expandedrichlistitems = addonsrichlist.
					getElementsByAttribute("type", "bandwagonAddonExpanded");
			if(expandedrichlistitems.length > 0)
			{
				let item = expandedrichlistitems[0];
				let vbox = item.boxObject.firstChild;
				vbox.appendChild(appendeditem);
			}
		},
		/**
		* Updates check box state according to whether currently displayed addon is
		* going into batch
		*/
		SetCheckBoxState: function()
		{
			if($eng.selectedcollection && $eng.selectedaddon && $ui._inclusion_checkbox)
			{
				let checked = !$eng.IsExcluded($eng.collection.bwurl,$eng.selectedaddon.guid);
				$ui._inclusion_checkbox.setAttribute("checked", checked);
			}
		},

		/**
		* Updates collection summary on the Batch Actions toolbar
		*/
		ReportSelection: function()
		{
			let colsummaryinfo = $dom.byid("massive-extender-summary-info");

			let menus =
			[
				$dom.byid("batch-actions-installation-menu"),
				$dom.byid("batch-actions-activation-menu"),
				$dom.byid("batch-actions-selection-menu")
			];
			let commandsformarked =
			[
				$dom.byid("massive-extender-install-selected-command"),
				$dom.byid("massive-extender-uninstall-selected-command"),
				$dom.byid("massive-extender-enable-selected-command"),
				$dom.byid("massive-extender-disable-selected-command")
			];
			let simplecommands =
			[
				$dom.byid("massive-extender-install-all-command"),
				$dom.byid("massive-extender-install-selected-command"),
				$dom.byid("massive-extender-mark-all-command"),
				$dom.byid("massive-extender-unmark-all-command")
			]

			simplecommands.forEach(function(command)
					(command.setAttribute("disabled", !($eng.selectedcollection && $eng.collection.addons))));

			if($eng.selectedcollection && $eng.collection.addons)
			{
				let selsize = $eng.collection.selectionsize;
				let colsize= $eng.collection.size;

				colsummaryinfo.value = "" + selsize + " / " + colsize;
				menus.forEach(function(menu) (menu.disabled = false));
				commandsformarked.forEach(function(command)
						(command.setAttribute("disabled", (selsize == 0))));
				$dom.byid("massive-extender-mark-all-command").
						setAttribute("disabled", (selsize == colsize));
				$dom.byid("massive-extender-unmark-all-command").
						setAttribute("disabled", (selsize == 0));
			}
			else
			{
				colsummaryinfo.value = "--";
				menus.forEach(function(menu) (menu.disabled = true));
			}
			
		},

		/**
		* Stylizes Add-On Collector add-on list to reflect user selection status
		* ([no-batch="true"] in CSS)
		*/
		StylizeBandwagonAddons: function()
		{
			if(!($eng.selectedcollection && $eng.collection.addons))
				return;
			let bwaddons = $dom.byid("bandwagon-addons-list").children;
			for each(let bwaddon in bwaddons)
			{
				if(bwaddon.addon) // TODO (low): this may be a redundant check since A.C. wouldn't display an empty item
				{
					let isexc = $eng.IsExcluded($eng.collection.bwurl, bwaddon.addon.guid);
					bwaddon.setAttribute("no-batch", isexc);
				}
			}
		},

		/**
		* Displays restart message after making changes to extensions registry
		*/
		UpdateRestartStatus: function(switchview)
		{
			if(switchview != "noviewswitch")
				showView("extensions");
			updateOptionalViews();
			if(gPendingActions)
				enableRestartButton();
		},
		
		UpdateViewOnLoad: function()
		{
			$ui.commands.SwitchUIMode();
		}
		
		
	} // massiveextender.App.Main.UI
	
	let $main=
	massiveextender.App.Main =
	{
		listeners:
		{
			ExtensionManager_load: function(ev)
			{
				$db.InitDB();
				$main.DoTheHacks();
				$ui.UpdateViewOnLoad();
			},
			ExtensionManager_unload: function(ev)
			{
				$db.CloseDB();
			},

			BWAddons_select: function(ev)
			{
				$ui.InsertInclusionCheckBox();
			}

		},
		
		
		DoTheHacks: function()
		{
		// ;; a spark of cruel Lisp cons weaving majiq...

			//
			// ...Bandwagon.Controller.CollectionsPane._repopulateAddonsList()
			//
			try
			{
				var __repopulate__unhacked__ =
						Bandwagon.Controller.CollectionsPane._repopulateAddonsList;
			}
			catch(ex)
			{
				// TODO (medium): degrade gracefully if A.C. is not initialized for some reason
			}
			Bandwagon.Controller.CollectionsPane._repopulateAddonsList =
			function(__collection__)
			{
				let __unhacked__result =
						__repopulate__unhacked__(__collection__);
				$ui.StylizeBandwagonAddons();
				$ui.ReportSelection();
				return __unhacked__result;
			}

			/* TODO (k.i.m.): we don't need this at the moment, but it's one of those thing
			  to check if an A.C. update breaks our modus operandi
			//
			// ...Bandwagon.Controller.CollectionsPane.doExpandAddon;
			//

			try
			{
				var __expandaddon__unhacked__ =
					Bandwagon.Controller.CollectionsPane.doExpandAddon;
			}
			catch(ex)
			{
				// TODO (medium): degrade gracefully if A.C. is not initialized for some reason
			}
			Bandwagon.Controller.CollectionsPane.doExpandAddon =
			function(__event__)
			{
				let __unhacked__result =
						$main.__expandaddon__unhacked__(__event__);
				$ui.InsertInclusionCheckBox();
				return __unhacked__result;
			}
			*/
			
		} // ...and they lived hackily ever after
	}
}
